/***************************************************************************
 *   Copyright (C) 2003 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the osgGtk library.                              *
 *                                                                         *
 *   The osgGtk library is free software; you can redistribute it and/or   *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The osgGtk library is distributed in the hope that it will be         *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/

// Of course, we need the gtk headers
#include <gtk/gtk.h>

// We'll need this to load the command line argument
// as well as the handling the file load button
#include <osgDB/ReadFile>

#include <osgGA/TrackballManipulator>

// And we'll need this for the Gtk OSG viewer
#include <osgGtk/ViewerGtk.h>

const char* HELP_TEXT =
    "<b><i>Right-click</i></b> while holding <b><i>CTRL</i></b> <u>or</u> <b><i>SHIFT</i></b> to pop up a fake Gtk menu.\n"
    "\n"
    "Use the standard OSG TrackballManipulator keys and mouse actions to\n"
    "rotate, zoom and move the loaded model.\n"
    "\n"
    "<i>Note:\n"
    "<small>\tIf the viewer is stopped, user actions won't be processed until the viewer is restarted.\n"
    "\tIn this example events are only processed when the viewer is running.</small></i>\n"
;

// This is global so that the callbacks can see it as well
// osgViewer::ViewerGtk* viewer = NULL;

// This is also global so the callback function can set the
// state of the toggle button
GtkWidget* fps_button;

// Callback to handle mouse button events
// This is what will look for right-clicks and pop up the menu
gboolean on_button_event(GtkWidget* widget, GdkEventButton* event, gpointer data);

// Callback that loads a selected scene as the root node of the scenegraph
void on_open_file_clicked(GtkWidget* widget, gpointer data);

// Callback that runs the viewer a specific number of frames
void on_run_frames_button_clicked(GtkWidget* widget, gpointer data);

// Callback that starts/stops the viewer
void on_fps_toggled(GtkWidget* widget, gpointer data);

int main( int argc, char** argv )
{
  // We'll set a basic size of 800x600, but the user can resize larger if desired
  // (but not smaller)
  int width=800, height=480;
  
  // This is some gtk housekeeping... we need the main loop and we also need to
  // initialize the OpenGL system
  gtk_init( &argc, &argv );
  gtk_gl_init( &argc, &argv );

  // Now, let's create our viewer using ViewerGtk so we can use the gtk
  // convenience method later
  //
  // We'll also set desired frame rate of 60 fps and load the first command line
  // argument as the scene
  osgViewer::ViewerGtk* viewer = new osgViewer::ViewerGtk();
  viewer->setCameraManipulator(new osgGA::TrackballManipulator);
  viewer->set_fps(60);

  if ( argc >= 2 ) {
    osg::ref_ptr<osg::Node> model = osgDB::readNodeFile(argv[1]);
    if ( model.valid() ) viewer->setSceneData(model.get());
  }

  // Now we'll setup the viewer in a gtk window using the setup_viewer_in_gtk_window()
  // convenience method
  //
  // The return type is a Gtk widget that we can use anywhere we want
  osgViewer::GraphicsWindowGtk* gw = viewer->setup_viewer_in_gtk_window(width,height);

  // All Gtk apps need a toplevel window
  //
  // We'll also set the title, add our widget created by the viewer and make sure
  // it's visible
  GtkWidget* window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), "osgviewerGtk");

  // Almost all of this is just Gtk code... creating the various
  // layouts and other UI widgets
  GtkWidget* main_vbox = gtk_vbox_new(FALSE, 3);
  GtkWidget* bottom_hbox = gtk_hbox_new(FALSE, 3);
  GtkWidget* button_vbox = gtk_vbox_new(TRUE, 3);
  GtkWidget* help_text = gtk_label_new(HELP_TEXT);
  GtkWidget* file_button = gtk_button_new_with_label("Open File");
  GtkWidget* run_frames_button = gtk_button_new_with_label("Run 120 frames");
  GtkWidget* close_button = gtk_button_new_with_label("Close");
  fps_button = gtk_toggle_button_new();

  gtk_label_set_use_markup(GTK_LABEL(help_text), TRUE);

  gtk_container_add(GTK_CONTAINER(window), main_vbox);
  gtk_container_add(GTK_CONTAINER(main_vbox), gw->gtk_widget());
  gtk_container_add(GTK_CONTAINER(main_vbox), bottom_hbox);
  gtk_container_add(GTK_CONTAINER(bottom_hbox), button_vbox);
  gtk_container_add(GTK_CONTAINER(bottom_hbox), help_text);
  gtk_container_add(GTK_CONTAINER(button_vbox), file_button);
  gtk_container_add(GTK_CONTAINER(button_vbox), run_frames_button);
  gtk_container_add(GTK_CONTAINER(button_vbox), fps_button);
  gtk_container_add(GTK_CONTAINER(button_vbox), close_button);
  gtk_widget_show_all(window);

  // Even though we won't do anything except popup the menu on button release
  // we still need to handle the case where the button is pressed with the
  // shift or control key. Otherwise the button press will be passed on to
  // the gtk handler.
  g_signal_connect( G_OBJECT(gw->gtk_widget()),
                    "button_press_event",
                    G_CALLBACK(on_button_event),
                    NULL
                  );
  g_signal_connect( G_OBJECT(gw->gtk_widget()),
                    "button_release_event",
                    G_CALLBACK(on_button_event),
                    NULL
                  );
  
  // Let's connect the other Gtk buttons to their callbacks
  g_signal_connect( G_OBJECT(file_button),
                    "clicked",
                    G_CALLBACK(on_open_file_clicked),
                    viewer
                  );
  g_signal_connect( G_OBJECT(run_frames_button),
                    "clicked",
                    G_CALLBACK(on_run_frames_button_clicked),
                    viewer
                  );
  g_signal_connect( G_OBJECT(fps_button),
                    "toggled",
                    G_CALLBACK(on_fps_toggled),
                    viewer
                  );
  g_signal_connect( G_OBJECT(close_button), "clicked", G_CALLBACK(gtk_main_quit), 0 );

  // And, just before we kick things off, let's start the viewer running
  // by setting the fps button to active, allowing the connected callback
  // to start things off.
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fps_button), TRUE);

  g_signal_connect(G_OBJECT(window), "delete_event", G_CALLBACK(gtk_main_quit), 0 );

  // This is boilerplate for all gtk apps... start the gtk main loop
  gtk_main();

  // If we got here, gtk exited it's main loop, so we'll exit cleanly
  return 0;
}

// This function is called every time there is a button press or button
// release event that occurs in the OSG graphics window
//
// Even though it is only the button release that we're concerned with,
// we still handle the button press event since anything that we don't
// handle is passed along to OSG. We don't want button press events
// to be passed into OSG when the user is holding down the shift
// key or the control key, so we stop that by returning true
// when either key is held down and the right mouse button is clicked.
gboolean on_button_event(GtkWidget* widget, GdkEventButton* event, gpointer data)
{
  if ( event->button == 3 and ( event->state & GDK_SHIFT_MASK or event->state & GDK_CONTROL_MASK ) )
  {
    // We get here if it's either a button press or release, but we'll
    // only pop up the menu on a release
    if ( event->type == GDK_BUTTON_RELEASE )
    {
      GtkWidget* menu = gtk_menu_new();
      gtk_menu_append(menu, gtk_menu_item_new_with_label("Option"));
      gtk_menu_append(menu, gtk_menu_item_new_with_label("Another Option"));
      gtk_menu_append(menu, gtk_menu_item_new_with_label("Still More Options"));
      gtk_widget_show_all(menu);
      gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);
    }

    // Whether it's press or release, if it's a right mouse event with
    // control or shift held down we return true, stopping it from
    // going on to OSG
    //
    // If we wanted to handled it _and_ let it pass into OSG we
    // could simply return false here.
    return true;
  }

  // But, since this was something other than a right mouse event with
  // a ctrl/shift key pressed, we'll just let it pass into OSG
  return false;
}

void on_open_file_clicked(GtkWidget* widget, gpointer data)
{
  osgViewer::ViewerGtk* viewer = (osgViewer::ViewerGtk*)data;
  GtkWidget* fc = gtk_file_chooser_dialog_new(
      "Please select an OSG file to load...",
      NULL,
      GTK_FILE_CHOOSER_ACTION_OPEN,
      GTK_STOCK_CANCEL,
      GTK_RESPONSE_CANCEL,
      GTK_STOCK_OPEN,
      GTK_RESPONSE_ACCEPT,
      NULL
  );

  if ( gtk_dialog_run(GTK_DIALOG(fc)) == GTK_RESPONSE_ACCEPT )
  {
    char* file = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(fc));
    osg::ref_ptr<osg::Node> model = osgDB::readNodeFile(file);
    if ( model.valid() ) viewer->setSceneData(model.get());
    g_free(file);
  }

  gtk_widget_destroy(fc);
}

void on_run_frames_button_clicked(GtkWidget* widget, gpointer data)
{
  osgViewer::ViewerGtk* viewer = (osgViewer::ViewerGtk*)data;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(fps_button), FALSE);
  if ( viewer ) viewer->run_frames(120);
}

void on_fps_toggled(GtkWidget* widget, gpointer data)
{
  osgViewer::ViewerGtk* viewer = (osgViewer::ViewerGtk*)data;
  if ( gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)))
  {
    if ( viewer ) viewer->run();
    gtk_button_set_label(GTK_BUTTON(widget), "Click to stop");
  }
  else
  {
    if ( viewer ) viewer->stop();
    gtk_button_set_label(GTK_BUTTON(widget), "Click to run at 60fps");
  }
}
