/***************************************************************************
 *   Copyright (C) 2003 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the osgGtk library.                              *
 *                                                                         *
 *   The osgGtk library is free software; you can redistribute it and/or   *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The osgGtk library is distributed in the hope that it will be         *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "GraphicsWindowGtk.h"

#include <iostream>
#include <gtk/gtkdrawingarea.h>

namespace osgViewer
{
  GraphicsWindowGtk::GraphicsWindowGtk(int width, int height, int mode):
      m_gtk_widget(NULL),
      m_gdk_gl_config(NULL)
  {
    // OSG housekeeping
    _traits = new GraphicsContext::Traits();
    _traits->x = 0;
    _traits->y = 0;
    _traits->width = width;
    _traits->height = height;

    this->setState( new osg::State );
    this->getState()->setGraphicsContext(this);

    if (_traits.valid() && _traits->sharedContext)
    {
      getState()->setContextID( _traits->sharedContext->getState()->getContextID() );
      incrementContextIDUsageCount( getState()->getContextID() );
    }
    else
    {
      getState()->setContextID( osg::GraphicsContext::createNewContextID() );
    }

    // GTK housekeeping
    m_gtk_widget = gtk_drawing_area_new();
    if ( m_gtk_widget == NULL ) throw(-1);

    gtk_widget_add_events( m_gtk_widget,
                           GDK_POINTER_MOTION_MASK |
                           GDK_BUTTON_MOTION_MASK |
                           GDK_BUTTON1_MOTION_MASK |
                           GDK_BUTTON2_MOTION_MASK |
                           GDK_BUTTON3_MOTION_MASK |
                           GDK_BUTTON_PRESS_MASK |
                           GDK_BUTTON_RELEASE_MASK |
                           GDK_KEY_PRESS_MASK |
                           GDK_KEY_RELEASE_MASK |
                           GDK_PROXIMITY_IN_MASK |
                           GDK_PROXIMITY_OUT_MASK
                         );

    m_gdk_gl_config = gdk_gl_config_new_by_mode( (GdkGLConfigMode) mode );
    if ( m_gdk_gl_config == NULL ) throw(-1);

    if ( ! gtk_widget_set_gl_capability( m_gtk_widget, m_gdk_gl_config, NULL, TRUE, GDK_GL_RGBA_TYPE ) ) throw(-1);

    this->getEventQueue()->setCurrentEventState(osgGA::GUIEventAdapter::getAccumulatedEventState().get());

    g_signal_connect(G_OBJECT(m_gtk_widget), "configure-event", G_CALLBACK(GraphicsWindowGtk::on_configure_event), this );

    g_signal_connect(G_OBJECT(m_gtk_widget), "button-press-event", G_CALLBACK(GraphicsWindowGtk::on_button_event), this );

    g_signal_connect(G_OBJECT(m_gtk_widget), "button-release-event", G_CALLBACK(GraphicsWindowGtk::on_button_event), this );

    g_signal_connect(G_OBJECT(m_gtk_widget), "key-press-event", G_CALLBACK(GraphicsWindowGtk::on_key_event), this );

    g_signal_connect(G_OBJECT(m_gtk_widget), "key-release-event", G_CALLBACK(GraphicsWindowGtk::on_key_event), this );

    g_signal_connect(G_OBJECT(m_gtk_widget), "motion-notify-event", G_CALLBACK(GraphicsWindowGtk::on_motion_notify_event), this );

    g_signal_connect(G_OBJECT(m_gtk_widget), "proximity-in-event", G_CALLBACK(GraphicsWindowGtk::on_proximity_event), this );

    g_signal_connect(G_OBJECT(m_gtk_widget), "proximity-out-event", G_CALLBACK(GraphicsWindowGtk::on_proximity_event), this );

    g_signal_connect(G_OBJECT(m_gtk_widget), "scroll-event", G_CALLBACK(GraphicsWindowGtk::on_scroll_event), this );

    g_object_set(m_gtk_widget, "can-focus", true, NULL);

    gtk_widget_set_size_request( m_gtk_widget, width, height );
  }

  GraphicsWindowGtk::~ GraphicsWindowGtk()
  {
  }

  bool GraphicsWindowGtk::isSameKindAs(const osg::Object* object) const
  {
    return dynamic_cast<const GraphicsWindowGtk*>(object)!=0;
  }

  const char* GraphicsWindowGtk::libraryName() const
  {
    return "osgGtk";
  }
  
  const char* GraphicsWindowGtk::className() const
  {
    return "GraphicsWindowGtk";
  }

  GtkWidget * GraphicsWindowGtk::gtk_widget()
  {
    return m_gtk_widget;
  }

  bool GraphicsWindowGtk::setWindowRectangleImplementation(int x, int y, int width, int height)
  {
    if ( not GTK_WIDGET_REALIZED(m_gtk_widget) ) return false;
    
    GtkRequisition req;
    req.width = width;
    req.height = height;
    gtk_widget_size_request(m_gtk_widget, &req);

    // from osgViewer::GraphicsWindowX11
    // add usleep here to give window manager a chance to handle the request, if
    // we don't add this sleep then any X11 calls right afterwards can produce
    // X11 errors.
    usleep(100000);
    
    return true;
  }

  bool GraphicsWindowGtk::setWindowDecorationImplementation(bool flag)
  {
    GdkWindow* root = gtk_widget_get_root_window(m_gtk_widget);

    if ( root == NULL ) return false;

    if ( flag )
      gdk_window_set_decorations( root, GDK_DECOR_ALL );
    else
      gdk_window_set_decorations( root, GDK_DECOR_BORDER );

    return true;
  }

  void GraphicsWindowGtk::setWindowName(const std::string & name)
  {
    if ( not GTK_WIDGET_REALIZED(m_gtk_widget) ) return;
    
    GdkWindow* root = gtk_widget_get_root_window(m_gtk_widget);
    if ( root == NULL ) return;
    gdk_window_set_title(root, name.c_str());
  }

  void GraphicsWindowGtk::grabFocus()
  {
    gtk_widget_grab_focus( m_gtk_widget );
  }

  void GraphicsWindowGtk::grabFocusIfPointerInWindow()
  {
    gint x, y;

    gtk_widget_get_pointer( m_gtk_widget, &x, &y );

    gint left   = m_gtk_widget->allocation.x,
         right  = m_gtk_widget->allocation.x + m_gtk_widget->allocation.width,
         top    = m_gtk_widget->allocation.y,
         bottom = m_gtk_widget->allocation.y + m_gtk_widget->allocation.height;

    if ( x >= left && x <= right && y >= top && y <= bottom )
      gtk_widget_grab_focus( m_gtk_widget );
  }

  bool GraphicsWindowGtk::valid() const
  {
    return true;
  }

  void GraphicsWindowGtk::setCursor(MouseCursor mouseCursor)
  {
    GdkCursor* cursor;

    switch(mouseCursor)
    {
      case InheritCursor:     cursor = NULL;                                    break;
      case NoCursor:          cursor = NULL;                                    break;
      case RightArrowCursor:  cursor = gdk_cursor_new(GDK_SB_RIGHT_ARROW);      break;
      case LeftArrowCursor:   cursor = gdk_cursor_new(GDK_SB_LEFT_ARROW);       break;
      case InfoCursor:        cursor = gdk_cursor_new(GDK_GUMBY);               break;
      case DestroyCursor:     cursor = gdk_cursor_new(GDK_CROSS_REVERSE);       break;
      case HelpCursor:        cursor = gdk_cursor_new(GDK_QUESTION_ARROW);      break;
      case CycleCursor:       cursor = gdk_cursor_new(GDK_EXCHANGE);            break;
      case SprayCursor:       cursor = gdk_cursor_new(GDK_SPRAYCAN);            break;
      case WaitCursor:        cursor = gdk_cursor_new(GDK_WATCH);               break;
      case TextCursor:        cursor = gdk_cursor_new(GDK_XTERM);               break;
      case CrosshairCursor:   cursor = gdk_cursor_new(GDK_TCROSS);              break;
      case UpDownCursor:      cursor = gdk_cursor_new(GDK_SB_V_DOUBLE_ARROW);   break;
      case LeftRightCursor:   cursor = gdk_cursor_new(GDK_SB_H_DOUBLE_ARROW);   break;
      case TopSideCursor:     cursor = gdk_cursor_new(GDK_TOP_SIDE);            break;
      case BottomSideCursor:  cursor = gdk_cursor_new(GDK_BOTTOM_SIDE);         break;
      case LeftSideCursor:    cursor = gdk_cursor_new(GDK_LEFT_SIDE);           break;
      case RightSideCursor:   cursor = gdk_cursor_new(GDK_RIGHT_SIDE);          break;
      case TopLeftCorner:     cursor = gdk_cursor_new(GDK_TOP_LEFT_CORNER);     break;
      case TopRightCorner:    cursor = gdk_cursor_new(GDK_TOP_RIGHT_CORNER);    break;
      case BottomRightCorner: cursor = gdk_cursor_new(GDK_BOTTOM_RIGHT_CORNER); break;
      case BottomLeftCorner:  cursor = gdk_cursor_new(GDK_BOTTOM_LEFT_CORNER);  break;
    }

    GdkWindow* root = gtk_widget_get_root_window( m_gtk_widget );
    if ( not root ) return;

    gdk_window_set_cursor(root, cursor);
    gdk_cursor_destroy(cursor);
  }

  bool GraphicsWindowGtk::realizeImplementation()
  {
    if ( GTK_WIDGET_REALIZED(m_gtk_widget) )
    {
      osg::notify(osg::NOTICE)<<"GraphicsWindowGtk::realizeImplementation() Already realized"<<std::endl;
      return true;
    }

    gtk_widget_realize(m_gtk_widget);
    
    return GTK_WIDGET_REALIZED(m_gtk_widget);
  }

  bool GraphicsWindowGtk::isRealizedImplementation() const
  {
    return GTK_WIDGET_REALIZED(m_gtk_widget);
  }

  bool GraphicsWindowGtk::makeCurrentImplementation()
  {
    if ( not GTK_WIDGET_REALIZED(m_gtk_widget) )
    {
      osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtk not realized, cannot do makeCurrent."<<std::endl;
      return false;
    }

    GdkGLDrawable* drawable = gtk_widget_get_gl_drawable(m_gtk_widget);
    GdkGLContext* context = gtk_widget_get_gl_context(m_gtk_widget);

    if ( not drawable or not context )
    {
      if ( not context )
        osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtk doesn't have a valid GL context, cannot do makeCurrent."<<std::endl;
      if ( not drawable )
        osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtk doesn't have a valid GL drawable, cannot do makeCurrent."<<std::endl;
      return false;
    }

    if ( not gdk_gl_drawable_make_current(drawable, context) )
    {
      osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtk couldn't call GL drawable's make_current(), cannot do makeCurrent."<<std::endl;
      return false;
    }
    
    return true;
  }

  bool GraphicsWindowGtk::releaseContextImplementation()
  {
    return true;
  }

  void GraphicsWindowGtk::closeImplementation()
  {

  }

  void GraphicsWindowGtk::swapBuffersImplementation()
  {
    this->swap_buffers();
  }

  bool GraphicsWindowGtk::swap_buffers()
  {
    GdkGLDrawable* drawable = gtk_widget_get_gl_drawable(m_gtk_widget);
    
    if ( not GTK_WIDGET_REALIZED(m_gtk_widget) or not drawable ) return false;

    if ( gdk_gl_drawable_is_double_buffered(drawable) )
      gdk_gl_drawable_swap_buffers(drawable);
    else
      glFlush();

    return true;
  }

  bool GraphicsWindowGtk::gl_begin()
  {
    GdkGLDrawable* drawable = gtk_widget_get_gl_drawable(m_gtk_widget);
    GdkGLContext* context = gtk_widget_get_gl_context(m_gtk_widget);

    if ( not GTK_WIDGET_REALIZED(m_gtk_widget) or not drawable or not context )
      return false;
    
    return gdk_gl_drawable_gl_begin( drawable, context );
  }

  bool GraphicsWindowGtk::gl_end()
  {
    GdkGLDrawable* drawable = gtk_widget_get_gl_drawable(m_gtk_widget);

    if ( not GTK_WIDGET_REALIZED(m_gtk_widget) or not drawable ) return false;
    
    gdk_gl_drawable_gl_end( drawable );

    return true;
  }

  gboolean GraphicsWindowGtk::on_button_event(GtkWidget * widget, GdkEventButton * event, gpointer user_data)
  {
    static bool emitting = false;
    bool handled;

    if (emitting) return false;

    emitting = true;
    handled = gtk_widget_event(widget, (GdkEvent*)event);
    emitting = false;

    if ( not handled and GTK_WIDGET_REALIZED(widget) )
    {
      GraphicsWindowGtk* gw = static_cast<GraphicsWindowGtk*>(user_data);
      double x = event->x;
      double y = event->y;
      gw->transform_mouse_x_y(x, y);
      
      switch (event->type)
      {
        case GDK_BUTTON_PRESS:
          gw->_eventQueue->mouseButtonPress( x, y, event->button );
          break;
        case GDK_2BUTTON_PRESS:
          gw->_eventQueue->mouseDoubleButtonPress( x, y, event->button );
          break;
        case GDK_BUTTON_RELEASE:
          gw->_eventQueue->mouseButtonRelease( x, y, event->button );
          break;
        case GDK_3BUTTON_PRESS: /* 3-button presses aren't handled in OSG */
        default:
          return false;
      }
    }
    
    // We always return true to indicate that we handled the signal
    // Otherwise the signal will emit a second time
    return TRUE;
  }

  gboolean GraphicsWindowGtk::on_configure_event(GtkWidget * widget, GdkEventConfigure * event, gpointer user_data)
  {
    GraphicsWindowGtk* gw = static_cast<GraphicsWindowGtk*>(user_data);
//     gw->resized(event->x, event->y, event->width, event->height);
    gw->_eventQueue->windowResize(event->x, event->y, event->width, event->height);
    return false;
  }

  gboolean GraphicsWindowGtk::on_key_event(GtkWidget * widget, GdkEventKey * event, gpointer user_data)
  {
    GraphicsWindowGtk* gw = static_cast<GraphicsWindowGtk*>(user_data);
    switch (event->type)
    {
      case GDK_KEY_PRESS_MASK:
        gw->_eventQueue->keyPress( event->keyval);
        break;
      case GDK_KEY_RELEASE_MASK:
        gw->_eventQueue->keyRelease( event->keyval);
        break;
      default:
        return false;
    }
    return true;
  }

  gboolean GraphicsWindowGtk::on_proximity_event(GtkWidget * widget, GdkEventProximity * event, gpointer user_data)
  {
    GraphicsWindowGtk* gw = static_cast<GraphicsWindowGtk*>(user_data);

    osgGA::GUIEventAdapter::TabletPointerType tab_type;

    switch (event->device->source)
    {
      case GDK_SOURCE_PEN:
        tab_type = osgGA::GUIEventAdapter::PEN;
        break;
      case GDK_SOURCE_ERASER:
        tab_type = osgGA::GUIEventAdapter::ERASER;
        break;
      case GDK_SOURCE_CURSOR:
        tab_type = osgGA::GUIEventAdapter::PUCK;
        break;
      default:
        tab_type = osgGA::GUIEventAdapter::UNKNOWN;
        break;
    }

    switch ( event->type )
    {
      case GDK_PROXIMITY_IN:
        gw->_eventQueue->penProximity(tab_type, true);
        break;
      case GDK_PROXIMITY_OUT:
        gw->_eventQueue->penProximity(tab_type, false);
        break;
      default:
        break;
    }
    return false;
  }

  gboolean GraphicsWindowGtk::on_motion_notify_event(GtkWidget * widget, GdkEventMotion * event, gpointer user_data)
  {
    GraphicsWindowGtk* gw = static_cast<GraphicsWindowGtk*>(user_data);
    gw->_eventQueue->mouseMotion(event->x, event->y);
    return true;
  }

  gboolean GraphicsWindowGtk::on_scroll_event(GtkWidget * widget, GdkEventScroll * event, gpointer user_data)
  {
    GraphicsWindowGtk* gw = static_cast<GraphicsWindowGtk*>(user_data);

    if ( event->device->num_axes >= 2 )
    {
      gw->_eventQueue->mouseScroll(osgGA::GUIEventAdapter::SCROLL_2D);
      gw->_eventQueue->mouseScroll2D(event->x, event->y);
    }
    else
    {
      osgGA::GUIEventAdapter::ScrollingMotion dir;
      switch(event->direction)
      {
        case GDK_SCROLL_UP:
          dir = osgGA::GUIEventAdapter::SCROLL_UP;
          break;
        case GDK_SCROLL_DOWN:
          dir = osgGA::GUIEventAdapter::SCROLL_DOWN;
          break;
        case GDK_SCROLL_LEFT:
          dir = osgGA::GUIEventAdapter::SCROLL_LEFT;
          break;
        case GDK_SCROLL_RIGHT:
          dir = osgGA::GUIEventAdapter::SCROLL_RIGHT;
          break;
      }
      gw->_eventQueue->mouseScroll(dir);
    }
    return false;
  }

  void GraphicsWindowGtk::transform_mouse_x_y(double & x, double & y)
  {
    // From osgViewer::GraphicsWindowX11::transformMouseXY
    if (this->getEventQueue()->getUseFixedMouseInputRange())
    {
      osgGA::GUIEventAdapter* eventState = this->getEventQueue()->getCurrentEventState();
      x = eventState->getXmin() + (eventState->getXmax()-eventState->getXmin())*x/double(_traits->width);
      y = eventState->getYmin() + (eventState->getYmax()-eventState->getYmin())*y/double(_traits->height);
    }

  }

}

