/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the osgGTK library.                              *
 *                                                                         *
 *   The osgGTK library is free software; you can redistribute it and/or   *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The osgGTK library is distributed in the hope that it will be         *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ViewerGtk.h"

#include <osgGA/TrackballManipulator>

namespace osgViewer
{

  ViewerGtk::ViewerGtk():
      Viewer(),
      m_is_running(false),
      m_frame_interval(30),
      m_run_priority(G_PRIORITY_DEFAULT),
      m_frame_step_rate(USE_REFERENCE_TIME),
      m_run_to_frame_number(0),
      m_timeout_source_id(0)
  {
  }

  ViewerGtk::ViewerGtk ( osg::ArgumentParser & arguments ) :
      Viewer ( arguments ),
      m_is_running(false),
      m_frame_interval(30),
      m_run_priority(G_PRIORITY_DEFAULT),
      m_frame_step_rate(USE_REFERENCE_TIME),
      m_run_to_frame_number(0),
      m_timeout_source_id(0)
  {
  }

  ViewerGtk::ViewerGtk ( const osgViewer::Viewer & viewer, const osg::CopyOp & copyop ) :
      Viewer ( viewer, copyop ),
      m_is_running(false),
      m_frame_interval(30),
      m_run_priority(G_PRIORITY_DEFAULT),
      m_frame_step_rate(USE_REFERENCE_TIME),
      m_run_to_frame_number(0),
      m_timeout_source_id(0)
  {
  }

  ViewerGtk::~ViewerGtk()
  {
    if ( this->is_running() ) this->stop();
  }

  bool ViewerGtk::isSameKindAs(const osg::Object* object) const
  {
    return dynamic_cast<const ViewerGtk*>(object)!=0;
  }

  const char* ViewerGtk::libraryName() const
  {
    return "osgGtk";
  }
  
  const char* ViewerGtk::className() const
  {
    return "ViewerGtk";
  }

  GraphicsWindowGtk * ViewerGtk::setup_viewer_in_gtk_window ( int width, int height )
  {
    GraphicsWindowGtk* gw = new GraphicsWindowGtk ( width, height );
    this->setThreadingModel ( osgViewer::Viewer::SingleThreaded );
    this->getCamera()->setViewport ( 0, 0, width, height );
    this->getCamera()->setProjectionMatrixAsPerspective ( 30.0f, static_cast<double> ( width ) /static_cast<double> ( height ), 1.0f, 10000.0f );
    this->getCamera()->setGraphicsContext ( gw );
    
    g_signal_connect(G_OBJECT(gw->gtk_widget()), "expose-event", G_CALLBACK(ViewerGtk::on_graphics_window_expose_event), this );

    return gw;
  }

  int ViewerGtk::run()
  {
    const char* str = getenv("OSG_RUN_FRAME_COUNT");
    if (str)
      m_run_to_frame_number = atoi(str);
    else
      m_run_to_frame_number = 0;

    return this->run_implementation();
  }

  int ViewerGtk::run_to_frame(int frame_num)
  {
    m_run_to_frame_number = frame_num;
    return this->run_implementation();
  }

  int ViewerGtk::run_frames(int frames)
  {
    if (frames <= 0 ) return 0;
    m_run_to_frame_number = getViewerFrameStamp()->getFrameNumber() + frames;
    return this->run_implementation();
  }

  void ViewerGtk::stop()
  {
    if ( m_timeout_source_id )
    {
      g_source_remove(m_timeout_source_id);
      m_timeout_source_id = 0;
    }

    m_is_running = false;
  }

  bool ViewerGtk::is_running()
  {
    return m_is_running;
  }

  double ViewerGtk::fps()
  {
    return 1000.0 / m_frame_interval;
  }

  void ViewerGtk::set_fps(double fps)
  {
    if ( m_frame_interval == 1000.0 / fps ) return;
    m_frame_interval = 1000.0 / fps;
    if ( this->is_running() ) this->connect_timer();
  }

  unsigned ViewerGtk::frame_interval_ms()
  {
    return m_frame_interval;
  }

  void ViewerGtk::set_frame_interval_ms(unsigned interval)
  {
    if ( m_frame_interval == interval ) return;
    m_frame_interval = interval;
    if ( this->is_running() ) this->connect_timer();
  }

  int ViewerGtk::run_priority()
  {
    return m_run_priority;
  }

  void ViewerGtk::set_run_priority(int priority)
  {
    if ( m_run_priority == priority ) return;
    m_run_priority = priority;
    if ( this->is_running() ) this->connect_timer();
  }

  double ViewerGtk::running_frame_step_rate()
  {
    return m_frame_step_rate;
  }

  void ViewerGtk::set_running_frame_step_rate(double frame_rate)
  {
    m_frame_step_rate = frame_rate;
  }

  int ViewerGtk::get_run_to_frame_number()
  {
    return m_run_to_frame_number;
  }

  gboolean ViewerGtk::on_graphics_window_expose_event(GtkWidget* widget, GdkEventExpose * event, gpointer data)
  {
    ViewerGtk* viewer = (ViewerGtk*)data;
    
    if (viewer->_done) return false;

    if (viewer->_firstFrame)
    {
      viewer->viewerInit();
        
      if (! viewer->isRealized())
      {
        viewer->realize();
      }
        
      viewer->_firstFrame = false;
    }

    // This is the only piece that is different from ViewerBase::frame()
//     viewer->advance(simulationTime);
    
    viewer->eventTraversal();
    viewer->updateTraversal();
    viewer->renderingTraversals();

    return false;
  }

  bool ViewerGtk::on_run_step()
  {
    // Should we stop or run one frame?
    if ( m_run_to_frame_number and getViewerFrameStamp()->getFrameNumber() >= m_run_to_frame_number)
    {
      this->stop();
      return 0;
    }
    else
    {
      // Step the frame
      this->frame(m_frame_step_rate);
    }

    return TRUE;
  }

  gboolean ViewerGtk::on_run_step_proxy(gpointer data)
  {
    if ( not data ) return FALSE;
    ViewerGtk* vgtk = (ViewerGtk*)data;
    return vgtk->on_run_step();
  }

  void ViewerGtk::connect_timer()
  {
    // Disconnect the old timer
    if ( m_timeout_source_id )
      g_source_remove(m_timeout_source_id);

    // Connect the new timer
    m_timeout_source_id = g_timeout_add_full( m_run_priority,
                                              m_frame_interval,
                                              on_run_step_proxy,
                                              this,
                                              NULL
                                            );
  }

  int ViewerGtk::run_implementation()
  {
    if (!getCameraManipulator() && getCamera()->getAllowEventFocus())
      setCameraManipulator(new osgGA::TrackballManipulator());
    
    if (!isRealized())
      realize();

    m_is_running = true;
      
    this->connect_timer();
      
    // If we didn't establish a connection to the timer, return an error according to ViewerBase::run() semantics
    if ( not m_timeout_source_id ) return -1;
    
    return 0;
  }

}


