/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the osgGTK library.                              *
 *                                                                         *
 *   The osgGTK library is free software; you can redistribute it and/or   *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The osgGTK library is distributed in the hope that it will be         *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef OSGVIEWERVIEWERGTK_H
#define OSGVIEWERVIEWERGTK_H

#include <osgViewer/Viewer>
#include <osgGtk/GraphicsWindowGtk.h>

/**
 * \mainpage osgGtk Library Reference
 *
 * The osgGtk library provides a framework allowing OpenSceneGraph
 * views to be incorporated into Gtk+ applications.
 *
 * \section dependencies Dependencies
 *  <table>
 *    <tr><td><b>Dependency</b></td><td><b>Version</b></td></tr>
 *    <tr><td><img src="osg-logo-small.png" alt="osg-logo"><br><a href="http://openscenegraph.org"><center>OpenSceneGraph</center></a></td><td>2.2 or higher</td></tr>
 *    <tr><td><img src="gtk-logo-small.png" alt="gtk-logo"><br><a href="http://gtk.org"><center>Gtk+</center></a></td><td>2.0 or higher</td></tr>
 *    <tr><td><a href="http://gtkglext.sourceforge.net">gtkglext</a></td><td>1.2 or higher</td></tr>
 *  </table>
 *
 * \section reference Important Reference Documentation Sections
 *  - <a href="annotated.html">Class List</a>
 *  - <a href="examples.html">All Examples</a>
 *
 * \section building Notes on building
 * A few notes on changes from the standard autotools (automake/autoconf) build process for osgGtk:
 *  - If you have a rpm build tree in @c ~/rpmbuild you can use @c make @c rpm to build a set of @c rpm files
 *    - In Fedora you can use @c rpmdev-setuptree to create a userspace rpm build tree
 *  - To build @c osgGtkmm you will need to add @c --enable-gtkmm when you run @c configure
 *  - To build the examples you will need to add @c --enable-examples when you run @c configure
 * 
 */

/**
 * @example osgviewerGtkmm/osgviewerGtkmm_simple.cpp
 *
 * Minimalistic example illustrating the how to use ViewerGtkmm and
 * GraphicsWindowGtkmm to display an OSG scenegraph in a Gtk-- application.
 */

/**
 * @example osgviewerGtkmm/osgviewerGtkmm.cpp
 *
 * This example illustrates how to use ViewerGtkmm and GraphicsWindowGtkmm to
 * display an OSG scenegraph in a Gtk-- application. Additionally, this
 * example incorporates several Gtk-- widgets to demonstrate several
 * aspects of ViewerGtkmm control as well as how to connect your own functions
 * to GraphicsWindowGtkmm mouse events.
 */

namespace osgViewer
{

  /**
   * Extends Viewer with several Gtk specific aspects.
   *
   * In particular, the ability to establish this viewer in a gtk graphics window with
   * the convenience method setup_viewer_in_gtk_window().
   *
   * Also reimplements run() to use the glib timer main loop
   *
   * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
   */
  class ViewerGtk : public Viewer
  {
    public:
      ViewerGtk();

      ViewerGtk ( osg::ArgumentParser& arguments );

      ViewerGtk ( const osgViewer::Viewer& viewer, const osg::CopyOp& copyop=osg::CopyOp::SHALLOW_COPY );

      /** If running, stops the viewer */
      virtual ~ViewerGtk();

      virtual bool isSameKindAs(const osg::Object* object) const;
      
      virtual const char* libraryName() const;

      virtual const char* className() const;

      /** Sets up this viewer to run in a Gtk window of the specified width and height */
      GraphicsWindowGtk* setup_viewer_in_gtk_window(int width, int height);

      /**
       * Starts the viewer running at the currently set frame interval and priority
       *
       * Runs until stopped
       */
      virtual int run();

      /** Runs the viewer until the viewer's OSG frame stamp reaches @a frame_num */
      int run_to_frame(int frame_num);

      /** Runs the viewer until the viewer's OSG frame stamp reaches the current frame stamp + @a frames */
      int run_frames(int frames);

      /** Stops the viewer from running */
      virtual void stop();

      /** True if the viewer is running, false otherwise */
      bool is_running();
      
      /** Returns the frames-per-second rate based on the currently set frame interval */
      double fps();

      /**
       * Sets the frames per second update rate
       *
       * Since the update interval has a resolution of 1ms, the update interval
       * may not be exactly the same as the set rate.
       *
       * For example, if set_fps(60) is called the actual update rate will not
       * be 16.66667 ms (1000 ms / 60). Instead, the update rate will be the
       * floor of 1000/60 or 16ms. Thus, the actual frame rate will be 62.5 fps.
       */
      void set_fps(double fps);

      /** Returns the frame interval in milliseconds that will be used when the viewer is running */
      unsigned frame_interval_ms();

      /** Sets the frame interval in milliseconds that will be used when the viewer is running */
      void set_frame_interval_ms(unsigned interval);

      /** Returns the run priority that will be used when the viewer is running */
      int run_priority();

      /** Sets the run priority that will be used when the viewer is running */
      void set_run_priority( int priority );

      /**
       * Returns the frame step rate that will be used when the viewer is running.
       * This is the value that will be passed to osgViewer::Viewer::advance() and
       * controls the rate at which the viewer is advanced.
       */
      double running_frame_step_rate();

      /** Sets the frame step rate that will be used when the viewer is running. */
      void set_running_frame_step_rate(double frame_rate);

      /** Returns the frame number that the viewer will run to */
      int get_run_to_frame_number();

    protected:
      bool m_is_running;
      unsigned m_frame_interval;
      int m_run_priority;
      double m_frame_step_rate;
      int m_run_to_frame_number;
      guint m_timeout_source_id;

      static gboolean on_graphics_window_expose_event(GtkWidget* widget, GdkEventExpose* event, gpointer data);

      virtual bool on_run_step();

      void connect_timer();

      virtual int run_implementation();

    private:

      /** Proxy callback point that passes the callback to the virtual on_run_step() */
      static gboolean on_run_step_proxy(gpointer data);
      
  };

}

#endif
