/***************************************************************************
 *   Copyright (C) 2003 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the osgGtk library.                              *
 *                                                                         *
 *   The osgGtk library is free software; you can redistribute it and/or   *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The osgGtk library is distributed in the hope that it will be         *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "GraphicsWindowGtkmm.h"

#include <gdkmm.h>

namespace osgViewer
{
  GraphicsWindowGtkmm::GraphicsWindowGtkmm(int width, int height, Gdk::GL::ConfigMode mode)
  {
    // OSG housekeeping
    _traits = new GraphicsContext::Traits();
    _traits->x = 0;
    _traits->y = 0;
    _traits->width = width;
    _traits->height = height;

    this->setState( new osg::State );
    this->getState()->setGraphicsContext(this);

    if (_traits.valid() && _traits->sharedContext)
    {
      getState()->setContextID( _traits->sharedContext->getState()->getContextID() );
      incrementContextIDUsageCount( getState()->getContextID() );
    }
    else
    {
      getState()->setContextID( osg::GraphicsContext::createNewContextID() );
    }

    // GTK housekeeping
    this->add_events( Gdk::POINTER_MOTION_MASK |
        Gdk::BUTTON_MOTION_MASK |
        Gdk::BUTTON1_MOTION_MASK |
        Gdk::BUTTON2_MOTION_MASK |
        Gdk::BUTTON3_MOTION_MASK |
        Gdk::BUTTON_PRESS_MASK |
        Gdk::BUTTON_RELEASE_MASK |
        Gdk::KEY_PRESS_MASK |
        Gdk::KEY_RELEASE_MASK |
        Gdk::PROXIMITY_IN_MASK |
        Gdk::PROXIMITY_OUT_MASK
                    );

    this->property_can_focus() = true;

    Glib::RefPtr<Gdk::GL::Config> glconfig;

    glconfig = Gdk::GL::Config::create( mode );

    if ( ! glconfig ) throw(-1);

    this->set_gl_capability( glconfig );

    this->getEventQueue()->setCurrentEventState(osgGA::GUIEventAdapter::getAccumulatedEventState().get());

    this->set_size_request( width, height );
  }

  GraphicsWindowGtkmm::~ GraphicsWindowGtkmm()
  {
  }

  bool GraphicsWindowGtkmm::isSameKindAs(const osg::Object* object) const
  {
    return dynamic_cast<const GraphicsWindowGtkmm*>(object)!=0;
  }

  const char* GraphicsWindowGtkmm::libraryName() const
  {
    return "osgGtkmm";
  }
  
  const char* GraphicsWindowGtkmm::className() const
  {
    return "GraphicsWindowGtkmm";
  }

  bool GraphicsWindowGtkmm::setWindowRectangleImplementation(int x, int y, int width, int height)
  {
    if ( not this->is_realized() ) return false;
    this->set_size_request( width, height );

    // from osgViewer::GraphicsWindowX11
    // add usleep here to give window manager a chance to handle the request, if
    // we don't add this sleep then any X11 calls right afterwards can produce
    // X11 errors.
    usleep(100000);
    
    return true;
  }

  bool GraphicsWindowGtkmm::setWindowDecorationImplementation(bool flag)
  {
    Glib::RefPtr<Gdk::Window> root = this->get_root_window();
    if ( not root ) return false;

    if ( flag )
      root->set_decorations( Gdk::DECOR_ALL );
    else
      root->set_decorations( Gdk::DECOR_BORDER );

    return true;
  }

  void GraphicsWindowGtkmm::setWindowName(const std::string & name)
  {
    if ( not this->is_realized() ) return;

    Glib::RefPtr<Gdk::Window> window = this->get_window();

    if ( not window ) return;

    window->set_title( name );
  }

  void GraphicsWindowGtkmm::grabFocus()
  {
    this->grab_focus();
  }

  void GraphicsWindowGtkmm::grabFocusIfPointerInWindow()
  {
    gint x, y;

    this->get_pointer( x, y );

    bool intersects;

    this->get_allocation().intersect(Gdk::Rectangle(x, y, 0, 0), intersects);

    if ( intersects ) this->grab_focus();
  }

  bool GraphicsWindowGtkmm::valid() const
  {
    return true;
  }

  void GraphicsWindowGtkmm::setCursor(MouseCursor mouseCursor)
  {
    Gdk::CursorType cursor;

    switch(mouseCursor)
    {
      case InheritCursor:     cursor = Gdk::X_CURSOR;            break;
      case NoCursor:          cursor = Gdk::X_CURSOR;            break;
      case RightArrowCursor:  cursor = Gdk::SB_RIGHT_ARROW;      break;
      case LeftArrowCursor:   cursor = Gdk::SB_LEFT_ARROW;       break;
      case InfoCursor:        cursor = Gdk::GUMBY;               break;
      case DestroyCursor:     cursor = Gdk::CROSS_REVERSE;       break;
      case HelpCursor:        cursor = Gdk::QUESTION_ARROW;      break;
      case CycleCursor:       cursor = Gdk::EXCHANGE;            break;
      case SprayCursor:       cursor = Gdk::SPRAYCAN;            break;
      case WaitCursor:        cursor = Gdk::WATCH;               break;
      case TextCursor:        cursor = Gdk::XTERM;               break;
      case CrosshairCursor:   cursor = Gdk::TCROSS;              break;
      case UpDownCursor:      cursor = Gdk::SB_V_DOUBLE_ARROW;   break;
      case LeftRightCursor:   cursor = Gdk::SB_H_DOUBLE_ARROW;   break;
      case TopSideCursor:     cursor = Gdk::TOP_SIDE;            break;
      case BottomSideCursor:  cursor = Gdk::BOTTOM_SIDE;         break;
      case LeftSideCursor:    cursor = Gdk::LEFT_SIDE;           break;
      case RightSideCursor:   cursor = Gdk::RIGHT_SIDE;          break;
      case TopLeftCorner:     cursor = Gdk::TOP_LEFT_CORNER;     break;
      case TopRightCorner:    cursor = Gdk::TOP_RIGHT_CORNER;    break;
      case BottomRightCorner: cursor = Gdk::BOTTOM_RIGHT_CORNER; break;
      case BottomLeftCorner:  cursor = Gdk::BOTTOM_LEFT_CORNER;  break;
    }

    Glib::RefPtr<Gdk::Window> root = this->get_root_window();
    if ( not root ) return;

    root->set_cursor( Gdk::Cursor(cursor) );
  }

  bool GraphicsWindowGtkmm::realizeImplementation()
  {
    if ( this->is_realized() )
    {
      osg::notify(osg::NOTICE)<<"GraphicsWindowGtkmm::realizeImplementation() Already realized"<<std::endl;
      return true;
    }

    this->Gtk::GL::DrawingArea::realize();

    return this->is_realized();
  }

  bool GraphicsWindowGtkmm::isRealizedImplementation() const
  {
    return this->is_realized();
  }

  bool GraphicsWindowGtkmm::makeCurrentImplementation()
  {
    if ( not this->is_realized() )
    {
      osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtkmm not realized, cannot do makeCurrent."<<std::endl;
      return false;
    }

    if ( not this->get_gl_context() or not this->get_gl_drawable() )
    {
      if ( not this->get_gl_context() )
        osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtkmm doesn't have a valid GL context, cannot do makeCurrent."<<std::endl;
      if ( not this->get_gl_drawable() )
        osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtkmm doesn't have a valid GL drawable, cannot do makeCurrent."<<std::endl;
      return false;
    }
    
    if ( not this->get_gl_drawable()->make_current( this->get_gl_context() ) )
    {
      osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtkmm couldn't call GL drawable's make_current(), cannot do makeCurrent."<<std::endl;
      return false;
    }

    return true;
  }

  bool GraphicsWindowGtkmm::releaseContextImplementation()
  {
//     if ( not this->is_realized() )
//     {
//       osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtkmm not realized, cannot do releaseContext."<<std::endl;
//       return false;
//     }
// 
//     if ( not this->get_gl_drawable() )
//     {
//       osg::notify(osg::NOTICE)<<"Warning: GraphicsWindowGtkmm does not have a GL drawable, cannot do releaseContext."<<std::endl;
//       return false;
//     }

    return true;
  }

  void GraphicsWindowGtkmm::closeImplementation()
  {

  }

  void GraphicsWindowGtkmm::swapBuffersImplementation()
  {
    this->swap_buffers();
  }

  bool GraphicsWindowGtkmm::swap_buffers()
  {
    if ( not this->is_realized() or not this->get_gl_drawable() ) return false;

    if ( this->get_gl_drawable()->is_double_buffered() )
      this->get_gl_drawable()->swap_buffers();
    else
      glFlush();

    return true;
  }

  bool GraphicsWindowGtkmm::gl_begin()
  {
    if ( not this->is_realized() or
         not this->get_gl_drawable() or
         not this->get_gl_context()
       ) return false;
    
    return this->get_gl_drawable()->gl_begin( this->get_gl_context() );
  }

  bool GraphicsWindowGtkmm::gl_end()
  {
    if ( not this->is_realized() or not this->get_gl_drawable() ) return false;
    
    this->get_gl_drawable()->gl_end();

    return true;
  }

  bool GraphicsWindowGtkmm::on_button_press_event(GdkEventButton * event)
  {
    static bool emitting = false;
    bool handled;

    if (emitting) return false;
    
    emitting = true;
    handled = this->event( (GdkEvent*)event );
    emitting = false;
    
    if ( not handled and this->is_realized() )
    {
      double x = event->x;
      double y = event->y;
      this->transform_mouse_x_y(x, y);
      
      switch (event->type)
      {
        case Gdk::BUTTON_PRESS:
          this->getEventQueue()->mouseButtonPress( x, y, event->button );
          break;
        case Gdk::DOUBLE_BUTTON_PRESS:
          this->getEventQueue()->mouseDoubleButtonPress( event->x, event->y, event->button );
          break;
        default:
          break;
      }
    }
    
    // We always return true to indicate that we handled the signal
    // Otherwise the signal will emit a second time
    return true;
  }

  bool GraphicsWindowGtkmm::on_button_release_event(GdkEventButton * event)
  {
    // We have to take the long way around since we want to check and
    // see if an event handler takes care of this event via a connected signal
    //
    // Otherwise gtkmm calls this method _then_ the connected signal slots
    //
    // But, if a connected signal handles the event we won't pass it
    // along to the OSG event queue
    static bool emitting = false;
    bool handled;

    if (emitting ) return false;
      
    emitting = true;
    handled = this->event( (GdkEvent*)event );
    emitting = false;

    if ( not handled and this->is_realized() )
    {
      double x = event->x;
      double y = event->y;
      this->transform_mouse_x_y(x, y);
      this->getEventQueue()->mouseButtonRelease( x, y, event->button );
    }

    // We always return true to indicate that we handled the signal
    // Otherwise the signal will emit a second time
    return true;
  }

  bool GraphicsWindowGtkmm::on_configure_event(GdkEventConfigure * event)
  {
    bool result = Gtk::GL::DrawingArea::on_configure_event(event);

    if ( _traits->x != event->x or _traits->y != event->y or
         _traits->width != event->width or _traits->height != event->height )
    {
      this->getEventQueue()->windowResize(event->x, event->y, event->width, event->height);
      this->resized(event->x, event->y, event->width, event->height);
    }
    
    return result;
  }

  bool GraphicsWindowGtkmm::on_key_press_event(GdkEventKey * event)
  {
    // We have to take the long way around since we want to check and
    // see if an event handler takes care of this event via a connected signal
    //
    // Otherwise gtkmm calls this method _then_ the connected signal slots
    //
    // But, if a connected signal handles the event we won't pass it
    // along to the OSG event queue
    static bool emitting = false;
    bool handled;

    if (emitting ) return false;
      
    emitting = true;
    handled = this->event( (GdkEvent*)event );
    emitting = false;

    if ( not handled and this->is_realized() )
      this->getEventQueue()->keyPress( event->keyval);

    // We always return true to indicate that we handled the signal
    // Otherwise the signal will emit a second time
    return true;
  }

  bool GraphicsWindowGtkmm::on_key_release_event(GdkEventKey * event)
  {
    // We have to take the long way around since we want to check and
    // see if an event handler takes care of this event via a connected signal
    //
    // Otherwise gtkmm calls this method _then_ the connected signal slots
    //
    // But, if a connected signal handles the event we won't pass it
    // along to the OSG event queue
    static bool emitting = false;
    bool handled;

    if (emitting ) return false;
      
    emitting = true;
    handled = this->event( (GdkEvent*)event );
    emitting = false;

    if ( not handled and this->is_realized() )
      this->getEventQueue()->keyRelease( event->keyval);

    // We always return true to indicate that we handled the signal
    // Otherwise the signal will emit a second time
    return true;
  }

  bool GraphicsWindowGtkmm::on_proximity_in_event(GdkEventProximity * event)
  {
    // We have to take the long way around since we want to check and
    // see if an event handler takes care of this event via a connected signal
    //
    // Otherwise gtkmm calls this method _then_ the connected signal slots
    //
    // But, if a connected signal handles the event we won't pass it
    // along to the OSG event queue
    static bool emitting = false;
    bool handled;

    if (emitting ) return false;
      
    emitting = true;
    handled = this->event( (GdkEvent*)event );
    emitting = false;

    if ( not handled and this->is_realized() )
    {

      osgGA::GUIEventAdapter::TabletPointerType tab_type;

      switch (event->device->source)
      {
        case Gdk::SOURCE_PEN:
          tab_type = osgGA::GUIEventAdapter::PEN;
          break;
        case Gdk::SOURCE_ERASER:
          tab_type = osgGA::GUIEventAdapter::ERASER;
          break;
        case Gdk::SOURCE_CURSOR:
          tab_type = osgGA::GUIEventAdapter::PUCK;
          break;
        default:
          tab_type = osgGA::GUIEventAdapter::UNKNOWN;
          break;
      }

      this->getEventQueue()->penProximity(tab_type, true);
    }

    // We always return true to indicate that we handled the signal
    // Otherwise the signal will emit a second time
    return true;
  }

  bool GraphicsWindowGtkmm::on_proximity_out_event(GdkEventProximity * event)
  {
    // We have to take the long way around since we want to check and
    // see if an event handler takes care of this event via a connected signal
    //
    // Otherwise gtkmm calls this method _then_ the connected signal slots
    //
    // But, if a connected signal handles the event we won't pass it
    // along to the OSG event queue
    static bool emitting = false;
    bool handled;

    if (emitting ) return false;
      
    emitting = true;
    handled = this->event( (GdkEvent*)event );
    emitting = false;

    if ( not handled and this->is_realized() )
    {
      osgGA::GUIEventAdapter::TabletPointerType tab_type;

      switch (event->device->source)
      {
        case Gdk::SOURCE_PEN:
          tab_type = osgGA::GUIEventAdapter::PEN;
          break;
        case Gdk::SOURCE_ERASER:
          tab_type = osgGA::GUIEventAdapter::ERASER;
          break;
        case Gdk::SOURCE_CURSOR:
          tab_type = osgGA::GUIEventAdapter::PUCK;
          break;
        default:
          tab_type = osgGA::GUIEventAdapter::UNKNOWN;
          break;
      }

      this->getEventQueue()->penProximity(tab_type, false);
    }
    
    // We always return true to indicate that we handled the signal
    // Otherwise the signal will emit a second time
    return true;
  }

  bool GraphicsWindowGtkmm::on_motion_notify_event(GdkEventMotion * event)
  {
    // We have to take the long way around since we want to check and
    // see if an event handler takes care of this event via a connected signal
    //
    // Otherwise gtkmm calls this method _then_ the connected signal slots
    //
    // But, if a connected signal handles the event we won't pass it
    // along to the OSG event queue
    static bool emitting = false;
    bool handled;

    if (emitting ) return false;
      
    emitting = true;
    handled = this->event( (GdkEvent*)event );
    emitting = false;

    if ( not handled and this->is_realized() )
    {
      double x = event->x;
      double y = event->y;
      this->transform_mouse_x_y(x, y);
      this->getEventQueue()->mouseMotion(x, y);
      return true;
    }

    // We always return true to indicate that we handled the signal
    // Otherwise the signal will emit a second time
    return true;
  }

  bool GraphicsWindowGtkmm::on_scroll_event(GdkEventScroll * event)
  {
    // We have to take the long way around since we want to check and
    // see if an event handler takes care of this event via a connected signal
    //
    // Otherwise gtkmm calls this method _then_ the connected signal slots
    //
    // But, if a connected signal handles the event we won't pass it
    // along to the OSG event queue
    static bool emitting = false;
    bool handled;

    if (emitting ) return false;
      
    emitting = true;
    handled = this->event( (GdkEvent*)event );
    emitting = false;

    if ( not handled and this->is_realized() )
    {
      if ( event->device->num_axes >= 2 )
      {
        this->getEventQueue()->mouseScroll(osgGA::GUIEventAdapter::SCROLL_2D);
        this->getEventQueue()->mouseScroll2D(event->x, event->y);
      }
      else
      {
        osgGA::GUIEventAdapter::ScrollingMotion dir;
        switch(event->direction)
        {
          case GDK_SCROLL_UP:
            dir = osgGA::GUIEventAdapter::SCROLL_UP;
            break;
          case GDK_SCROLL_DOWN:
            dir = osgGA::GUIEventAdapter::SCROLL_DOWN;
            break;
          case GDK_SCROLL_LEFT:
            dir = osgGA::GUIEventAdapter::SCROLL_LEFT;
            break;
          case GDK_SCROLL_RIGHT:
            dir = osgGA::GUIEventAdapter::SCROLL_RIGHT;
            break;
        }
        this->getEventQueue()->mouseScroll(dir);
      }
    }

    // We always return true to indicate that we handled the signal
    // Otherwise the signal will emit a second time
    return true;
  }

  void GraphicsWindowGtkmm::transform_mouse_x_y(double & x, double & y)
  {
    // From osgViewer::GraphicsWindowX11::transformMouseXY
    if (this->getEventQueue()->getUseFixedMouseInputRange())
    {
      osgGA::GUIEventAdapter* eventState = this->getEventQueue()->getCurrentEventState();
      x = eventState->getXmin() + (eventState->getXmax()-eventState->getXmin())*x/double(_traits->width);
      y = eventState->getYmin() + (eventState->getYmax()-eventState->getYmin())*y/double(_traits->height);
    }

  }

}

