/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the osgGTK library.                              *
 *                                                                         *
 *   The osgGTK library is free software; you can redistribute it and/or   *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The osgGTK library is distributed in the hope that it will be         *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ViewerGtkmm.h"

#include <osgGA/TrackballManipulator>

namespace osgViewer
{

  ViewerGtkmm::ViewerGtkmm():
      Viewer(),
      m_is_running(false),
      m_frame_interval(30),
      m_run_priority(Glib::PRIORITY_DEFAULT),
      m_frame_step_rate(USE_REFERENCE_TIME),
      m_run_to_frame_number(0)
  {
  }

  ViewerGtkmm::ViewerGtkmm ( osg::ArgumentParser & arguments ) :
      Viewer ( arguments ),
      m_is_running(false),
      m_frame_interval(30),
      m_run_priority(Glib::PRIORITY_DEFAULT),
      m_frame_step_rate(USE_REFERENCE_TIME),
      m_run_to_frame_number(0)
  {
  }

  ViewerGtkmm::ViewerGtkmm ( const osgViewer::Viewer & viewer, const osg::CopyOp & copyop ) :
      Viewer ( viewer, copyop ),
      m_is_running(false),
      m_frame_interval(30),
      m_run_priority(Glib::PRIORITY_DEFAULT),
      m_frame_step_rate(USE_REFERENCE_TIME),
      m_run_to_frame_number(0)
  {
  }

  ViewerGtkmm::~ViewerGtkmm()
  {
    if ( this->is_running() ) this->stop();
  }

  bool ViewerGtkmm::isSameKindAs(const osg::Object* object) const
  {
    return dynamic_cast<const ViewerGtkmm*>(object)!=0;
  }

  const char* ViewerGtkmm::libraryName() const
  {
    return "osgGtk";
  }
  
  const char* ViewerGtkmm::className() const
  {
    return "ViewerGtkmm";
  }

  GraphicsWindowGtkmm * ViewerGtkmm::setup_viewer_in_gtkmm_window ( int width, int height )
  {
    GraphicsWindowGtkmm* gw = new GraphicsWindowGtkmm ( width, height );
    this->setThreadingModel ( osgViewer::Viewer::SingleThreaded );
    this->getCamera()->setViewport ( 0, 0, width, height );
    this->getCamera()->setProjectionMatrixAsPerspective ( 30.0f, static_cast<double> ( width ) /static_cast<double> ( height ), 1.0f, 10000.0f );
    this->getCamera()->setGraphicsContext ( gw );
    gw->signal_expose_event().connect(sigc::mem_fun(*this, &ViewerGtkmm::on_graphics_window_expose_event));
    return gw;
  }

  int ViewerGtkmm::run()
  {
    const char* str = getenv("OSG_RUN_FRAME_COUNT");
    if (str)
      m_run_to_frame_number = atoi(str);
    else
      m_run_to_frame_number = 0;

    return this->run_implementation();
  }

  int ViewerGtkmm::run_to_frame(int frame_num)
  {
    m_run_to_frame_number = frame_num;
    return this->run_implementation();
  }

  int ViewerGtkmm::run_frames(int frames)
  {
    if (frames <= 0 ) return 0;
    m_run_to_frame_number = getViewerFrameStamp()->getFrameNumber() + frames;
    return this->run_implementation();
  }

  void ViewerGtkmm::stop()
  {
    if ( m_timer_connection ) m_timer_connection.disconnect();

    m_is_running = false;
  }

  bool ViewerGtkmm::is_running()
  {
    return m_is_running;
  }

  double ViewerGtkmm::fps()
  {
    return 1000.0 / m_frame_interval;
  }

  void ViewerGtkmm::set_fps(double fps)
  {
    if ( m_frame_interval == 1000.0 / fps ) return;
    m_frame_interval = 1000.0 / fps;
    if ( this->is_running() ) this->connect_timer();
  }

  unsigned ViewerGtkmm::frame_interval_ms()
  {
    return m_frame_interval;
  }

  void ViewerGtkmm::set_frame_interval_ms(unsigned interval)
  {
    if ( m_frame_interval == interval ) return;
    m_frame_interval = interval;
    if ( this->is_running() ) this->connect_timer();
  }

  int ViewerGtkmm::run_priority()
  {
    return m_run_priority;
  }

  void ViewerGtkmm::set_run_priority(int priority)
  {
    if ( m_run_priority == priority ) return;
    m_run_priority = priority;
    if ( this->is_running() ) this->connect_timer();
  }

  double ViewerGtkmm::running_frame_step_rate()
  {
    return m_frame_step_rate;
  }

  void ViewerGtkmm::set_running_frame_step_rate(double frame_rate)
  {
    m_frame_step_rate = frame_rate;
  }

  int ViewerGtkmm::get_run_to_frame_number()
  {
    return m_run_to_frame_number;
  }

  bool ViewerGtkmm::on_graphics_window_expose_event(GdkEventExpose * event)
  {
    if (_done) return false;

    if (_firstFrame)
    {
      viewerInit();
        
      if (!isRealized())
      {
        realize();
      }
        
      _firstFrame = false;
    }

    // This is the only piece that is different from ViewerBase::frame()
//     advance(simulationTime);
    
    eventTraversal();
    updateTraversal();
    renderingTraversals();

    return false;
  }

  int ViewerGtkmm::on_run_step()
  {
    // Should we stop or run one frame?
    if ( m_run_to_frame_number and getViewerFrameStamp()->getFrameNumber() >= m_run_to_frame_number)
    {
      this->stop();
      return 0;
    }
    else
    {
      // Step the frame
      this->frame(m_frame_step_rate);
    }

    return 1;
  }

  int ViewerGtkmm::on_run_step_proxy()
  {
    return this->on_run_step();
  }

  void ViewerGtkmm::connect_timer()
  {
    // Disconnect the old timer
    if ( m_timer_connection ) m_timer_connection.disconnect();

    // Connect the new timer
    m_timer_connection = Glib::signal_timeout().connect(sigc::mem_fun(*this, &ViewerGtkmm::on_run_step_proxy), m_frame_interval, m_run_priority);
  }

  int ViewerGtkmm::run_implementation()
  {
    if (!getCameraManipulator() && getCamera()->getAllowEventFocus())
      setCameraManipulator(new osgGA::TrackballManipulator());
    
    if (!isRealized())
      realize();

    m_is_running = true;
      
    this->connect_timer();
      
    // If we didn't establish a connection to the timer, return an error according to ViewerBase::run() semantics
    if ( not m_timer_connection ) return -1;
    
    return 0;
  }

}


