/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the osgGTK library.                              *
 *                                                                         *
 *   The osgGTK library is free software; you can redistribute it and/or   *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The osgGTK library is distributed in the hope that it will be         *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef OSGVIEWERVIEWERGTKMM_H
#define OSGVIEWERVIEWERGTKMM_H

#include <osgViewer/Viewer>
#include <osgGtkmm/GraphicsWindowGtkmm.h>

/**
 * \mainpage osgGtkmm Library Reference
 *
 * The osgGtkmm library provides a framework allowing OpenSceneGraph
 * views to be incorporated into Gtk-- applications.
 *
 * \section dependencies Dependencies
 * <table>
 *  <tr><td><b>Dependency</b></td><td><b>Version</b></td></tr>
 *  <tr><td><img src="osg-logo-small.png" alt="osg-logo"><br><a href="http://openscenegraph.org"><center>OpenSceneGraph</center></a></td><td>2.2 or higher</td></tr>
 *  <tr><td><img src="gtk-logo-small.png" alt="gtk-logo"><br><a href="http://gtkmm.org"><center>Gtk--</center></a></td><td>2.6 or higher</td></tr>
 *  <tr><td><a href="http://gtkglext.sourceforge.net">gtkglextmm</a></td><td>1.2 or higher</td></tr>
 * </table>
 *
 * \section osgsamples Sample OSG Datasets
 * The sample applications load a single file and display it as the root scenegraph node.
 *
 * Sample files can be found here:
 * <a href="http://www.openscenegraph.org/projects/osg/wiki/Downloads/SampleDatasets">http://www.openscenegraph.org/projects/osg/wiki/Downloads/SampleDatasets</a>
 * 
 * When testing, I most often use @c spaceship.osg or @c fountain.osg (found in the OpenSceneGraph-Data zip file
 * on the aforementioned page) since they are animated nodes that provide the best feedback as to when the viewer
 * is running or stopped.
 * 
 * \section reference Important Reference Documentation Sections
 *  - <a href="annotated.html">Class List</a>
 *  - <a href="examples.html">All Examples</a>
 * 
 * \section building Notes on building
 * A few notes on changes from the standard autotools (automake/autoconf) build process for osgGtk:
 *  - If you have a rpm build tree in @c ~/rpmbuild you can use @c make @c rpm to build a set of @c rpm files
 *    - In Fedora you can use @c rpmdev-setuptree to create a userspace rpm build tree
 *  - To build @c osgGtkmm you will need to add @c --enable-gtkmm when you run @c configure
 *  - To build the examples you will need to add @c --enable-examples when you run @c configure
 * 
 */

/**
 * @example osgviewerGtkmm/osgviewerGtkmm_simple.cpp
 *
 * Minimalistic example illustrating the how to use ViewerGtkmm and
 * GraphicsWindowGtkmm to display an OSG scenegraph in a Gtk-- application.
 */

/**
 * @example osgviewerGtkmm/osgviewerGtkmm.cpp
 *
 * This example illustrates how to use ViewerGtkmm and GraphicsWindowGtkmm to
 * display an OSG scenegraph in a Gtk-- application. Additionally, this
 * example incorporates several Gtk-- widgets to demonstrate several
 * aspects of ViewerGtkmm control as well as how to connect your own functions
 * to GraphicsWindowGtkmm mouse events.
 */

namespace osgViewer
{

  /**
   * Extends OSG's Viewer with several Gtkmm specific aspects.
   *
   * In particular, the ability to establish this viewer in a gtkmm graphics window with
   * the convenience method setup_viewer_in_gtkmm_window().
   *
   * Also reimplements run() to use the glibmm timer main loop
   *
   * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
  */
  class ViewerGtkmm : public Viewer
  {
    public:
      ViewerGtkmm();

      ViewerGtkmm ( osg::ArgumentParser& arguments );

      ViewerGtkmm ( const osgViewer::Viewer& viewer, const osg::CopyOp& copyop=osg::CopyOp::SHALLOW_COPY );

      /** If running, stops the viewer */
      virtual ~ViewerGtkmm();

      virtual bool isSameKindAs(const osg::Object* object) const;
      
      virtual const char* libraryName() const;

      virtual const char* className() const;

      /** Sets up this viewer to run in a Gtkmm window of the specified width and height */
      GraphicsWindowGtkmm* setup_viewer_in_gtkmm_window(int width, int height);

      /**
       * Starts the viewer running at the currently set frame interval and priority
       *
       * Runs until stopped
       */
      virtual int run();

      /** Runs the viewer until the viewer's OSG frame stamp reaches @a frame_num */
      int run_to_frame(int frame_num);

      /** Runs the viewer until the viewer's OSG frame stamp reaches the current frame stamp + @a frames */
      int run_frames(int frames);

      /** Stops the viewer from running */
      virtual void stop();

      /** True if the viewer is running, false otherwise */
      bool is_running();

      /** Returns the frames-per-second rate based on the currently set frame interval */
      double fps();

      /**
       * Sets the frames per second update rate
       *
       * Since the update interval has a resolution of 1ms, the update interval
       * may not be exactly the same as the set rate.
       *
       * For example, if set_fps(60) is called the actual update rate will not
       * be 16.66667 ms (1000 ms / 60). Instead, the update rate will be the
       * floor of 1000/60 or 16ms. Thus, the actual frame rate will be 62.5 fps.
       */
      void set_fps(double fps);

      /** Returns the frame interval in milliseconds that will be used when the viewer is running */
      unsigned frame_interval_ms();

      /** Sets the frame interval in milliseconds that will be used when the viewer is running */
      void set_frame_interval_ms(unsigned interval);

      /** Returns the run priority that will be used when the viewer is running */
      int run_priority();

      /** Sets the run priority that will be used when the viewer is running */
      void set_run_priority( int priority );

      /**
       * Returns the frame step rate that will be used when the viewer is running.
       * This is the value that will be passed to osgViewer::Viewer::advance() and
       * controls the rate at which the viewer is advanced.
       */
      double running_frame_step_rate();

      /** Sets the frame step rate that will be used when the viewer is running. */
      void set_running_frame_step_rate(double frame_rate);

      /** Returns the frame number that the viewer will run to */
      int get_run_to_frame_number();

    protected:
      bool m_is_running;
      unsigned m_frame_interval;
      int m_run_priority;
      double m_frame_step_rate;
      int m_run_to_frame_number;
      sigc::connection m_timer_connection;

      bool on_graphics_window_expose_event(GdkEventExpose* event);

      virtual int on_run_step();

      void connect_timer();

      virtual int run_implementation();

    private:

      /** Proxy callback point that passes the callback to the virtual on_run_step() */
      int on_run_step_proxy();
      
  };

}

#endif
